/*****************************************************************************
 * cuckooo - OpenOffice.org nested in KDE                                    *
 *                                                                           *
 * Copyright (C) 2003 Jan Holesovsky <kendy@artax.karlin.mff.cuni.cz>        *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License as published by      *
 * the Free Software Foundation; either version 2 of the License, or         *
 * (at your option) any later version.                                       *
 *                                                                           *
 * This program is distributed in the hope that it will be useful,           *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 * GNU General Public License for more details.                              *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to the Free Software               *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *****************************************************************************/

#include <unistd.h>

#include <qpainter.h>

#include <kapplication.h>
#include <kdebug.h>
#include <kstandarddirs.h>

#include <com/sun/star/awt/XSystemChildFactory.hpp>
#include <com/sun/star/awt/XToolkit.hpp>
#include <com/sun/star/beans/PropertyValue.hpp>
#include <com/sun/star/frame/FrameSearchFlag.hpp>
#include <com/sun/star/frame/XComponentLoader.hpp>
#include <com/sun/star/frame/XDispatch.hpp>
#include <com/sun/star/frame/XDispatchProvider.hpp>
#include <com/sun/star/frame/XFrame.hpp>
#include <com/sun/star/lang/SystemDependent.hpp>
#include <com/sun/star/lang/XMultiServiceFactory.hpp>
#include <com/sun/star/util/URL.hpp>
#include <com/sun/star/util/XURLTransformer.hpp>
#include <cppuhelper/bootstrap.hxx>
#include <rtl/ustrbuf.hxx>
#include <rtl/string.hxx>

#include <sfxsids.hrc>

#include "officewidget.h"

using namespace com::sun::star;
using namespace cppu;
using namespace rtl;

using com::sun::star::uno::Any;
using com::sun::star::uno::Exception;
using com::sun::star::uno::Reference;
using com::sun::star::uno::Sequence;

using namespace KOOo;

#define OUSTR(x) ::rtl::OUString( RTL_CONSTASCII_USTRINGPARAM(x) )

QString OUString2Q(const OUString &str)
{
  QChar *uni = new QChar[str.getLength() + 1];
  
  const sal_Unicode *ouPtr = str.getStr();
  const sal_Unicode *ouEnd = str.getStr() + str.getLength();
  QChar *qPtr = uni;

  while (ouPtr != ouEnd)
    *(qPtr++) = *(ouPtr++);

  *qPtr = 0;
  
  QString ret(uni, str.getLength());
  delete[] uni;

  return ret;
}

OUString QString2OU(const QString &str)
{
  sal_Unicode *uni = new sal_Unicode[str.length() + 1];
  
  const QChar *qPtr = str.unicode();
  const QChar *qEnd = str.unicode() + str.length();
  sal_Unicode *uPtr = uni;

  while (qPtr != qEnd)
    *(uPtr++) = *(qPtr++);

  *uPtr = 0;
  
  OUString ret(uni, str.length());
  delete[] uni;

  return ret;
}

QPixmap *OfficeWidget::s_splashPixmap = NULL;

OfficeWidget::OfficeWidget(const QString &conspec,
    const QString &officefname, const QString &inifile,
    QWidget *parent, const char *name, WFlags f)
  : QWidget(parent, name, f),
    m_officeProcess(NULL),
    m_unoUrlResolver(),
    m_splashProgress(100),
    m_frame()
{
  m_conSpec = QString2OU(conspec);

  initOffice(QString2OU(officefname), QString2OU(inifile));

  if (!s_splashPixmap)
    s_splashPixmap = new QPixmap(locate("data", "cuckooo/intro.png"));

  setFocusPolicy(QWidget::ClickFocus);
}

OfficeWidget::~OfficeWidget(void)
{
  Reference<lang::XMultiServiceFactory> rServiceManager = serviceManager(1);
  if (rServiceManager.is())
  {
    if (m_frame.is())
      m_frame->dispose();

    if (m_officeProcess)
      osl_terminateProcess(m_officeProcess);
  }
}

bool OfficeWidget::openFile(QString &name)
{
  return loadDocument(QString2OU(name));
}

bool OfficeWidget::closeFile(void)
{
  if (!m_frame.is())
    return true;
 
  // TODO: Just close the document!

  return false;
}

void OfficeWidget::setSplashProgress(int progress)
{
  if (progress == m_splashProgress)
    return;
  
  int old_progress = m_splashProgress;
  m_splashProgress = progress;

  // repaint the splashscreen
  if ((old_progress <  100 && m_splashProgress >= 100) ||
      (old_progress >= 100 && m_splashProgress <  100))
    repaint(true);
  else
    repaint(false);
}

void OfficeWidget::paintEvent(QPaintEvent *)
{
  if (100 <= m_splashProgress)
    return;

  QPainter paint(this);
  
  int origin_x = (width() - s_splashPixmap->width())/2;
  int origin_y = (height() - s_splashPixmap->height())/2;

  // draw splash pixmap
  paint.setPen(QPen(QColor(0, 0, 0), 0, Qt::SolidLine));
  paint.drawPixmap(origin_x, origin_y, *s_splashPixmap);

  if (0 <= m_splashProgress)
  {
    const int progbar_space  = 10;
    const int prog_space     = 2;
    const int prog_height    = 4;

    int progbar_origin_x = origin_x + progbar_space;
    int progbar_origin_y = origin_y + s_splashPixmap->height()
      - progbar_space - 2*prog_space - prog_height;
    int prog_full = s_splashPixmap->width() - 2*progbar_space - 2*prog_space;
    
    // draw progressbar
    paint.setPen(Qt::NoPen);
    paint.setBrush(QColor(250, 250, 250));
    paint.drawRect(progbar_origin_x, progbar_origin_y,
	s_splashPixmap->width() - 2*progbar_space, 2*prog_space + prog_height);

    paint.setBrush(QColor(16, 48, 144));
    paint.drawRect(progbar_origin_x + prog_space,
	progbar_origin_y + prog_space,
	(m_splashProgress*prog_full)/100, prog_height);
  }
}

bool OfficeWidget::initOffice(const rtl::OUString &officefname, const OUString &inifile)
{
  // Create the initial component context
  Reference<uno::XComponentContext> rComponentContext;
  try
  {
    if (inifile.getLength() > 0)
      rComponentContext = defaultBootstrap_InitialComponentContext(inifile);
    else
      rComponentContext = defaultBootstrap_InitialComponentContext();
  }
  catch (Exception &e)
  {
    OString o = OUStringToOString(e.Message, RTL_TEXTENCODING_ASCII_US);
    kdDebug() << "Unable to get initial component context: " << o.pData->buffer << endl;
    return false;
  }
  catch (...)
  {
    kdDebug() << "Unable to get initial component context." << endl;
    return false;
  }

  // Get the (local) servicemanager from the context
  Reference<lang::XMultiComponentFactory> rLocalServiceManager =
    rComponentContext->getServiceManager();

  // Instantiate the url resolver
  Reference<uno::XInterface> rInstance =
    rLocalServiceManager->createInstanceWithContext(OUSTR("com.sun.star.bridge.UnoUrlResolver"),
	rComponentContext);

  m_unoUrlResolver = Reference<bridge::XUnoUrlResolver>(rInstance, uno::UNO_QUERY);
  if (!m_unoUrlResolver.is())
  {
    kdDebug() << "Couldn't instantiate com.sun.star.bridge.UnoUrlResolver "
                 "service." << endl;
    return false;
  }

  // Try once to get the service manager
  Reference<lang::XMultiServiceFactory> rServiceManager = serviceManager(1);
  if (!rServiceManager.is())
  {
    // OOo is most probably not running, start it (or make it listening).
    OUString ouArg1 = OUSTR("-accept=") + m_conSpec + OUSTR(";urp;StarOffice.ServiceManager");
    OUString ouArg2 = OUSTR("-headless"); // no UI at all
    const int NO_PARGS = 2;
    rtl_uString *pArgs[NO_PARGS];
    pArgs[0] = ouArg1.pData;
    pArgs[1] = ouArg2.pData;

    // search the $PATH for the program and detach the process
    oslProcessOption processOptions = (osl_Process_SEARCHPATH | osl_Process_DETACHED);

    // run OpenOffice.org
    oslProcessError error = osl_executeProcess(officefname.pData, pArgs, NO_PARGS,
	processOptions, 0, NULL, NULL, 0, &m_officeProcess);

    if (error != osl_Process_E_None)
    {
      kdError() << "Unable to start OpenOffice.org. Error: " << error << endl;
      return false;
    }
  }

  return true;
}

Reference<lang::XMultiServiceFactory> OfficeWidget::serviceManager(int notimes, int delay)
{
  if (!m_unoUrlResolver.is())
  {
    kdError() << "UnoUrlResolver is not valid." << endl;
    return Reference<lang::XMultiServiceFactory>();
  }
  
  OUString ouConArg = OUSTR("uno:") + m_conSpec + OUSTR(";urp;StarOffice.ServiceManager");

  int countdown = notimes;
  while (countdown--)
  {
    try
    {
      Reference<uno::XInterface> rInstance = m_unoUrlResolver->resolve(ouConArg);
      if (!rInstance.is())
      {
	kdDebug() << "StarOffice.ServiceManager is not exported from the "
	  "remote counterpart." << endl;
	return Reference<lang::XMultiServiceFactory>();
      }

      // query for the simpler XMultiServiceFactory interface, sufficient for scripting
      Reference<lang::XMultiServiceFactory> rServiceManager(rInstance, uno::UNO_QUERY);
      if (rServiceManager.is())
      {
	setSplashProgress(100);
	return rServiceManager;
      }
    }
    catch (Exception &e)
    {
      OString o = OUStringToOString(e.Message, RTL_TEXTENCODING_ASCII_US);
      kdDebug() << "Connection to OOo failed: " << o.pData->buffer << endl;
    }
    catch (...)
    {
      kdDebug() << "Connection to OOo failed." << endl;
    }
    
    // Wait unless it is the last try
    if (countdown)
    {
      kdDebug() << "Waiting for OOo to start up: will try " << countdown
	        << " more times." << endl;

      // FIXME: real progress here, this is just a fake
      setSplashProgress((100*(notimes - countdown))/notimes);

      KApplication::kApplication()->processEvents();
      sleep(delay);
    }
  }
  
  setSplashProgress(100);
  return Reference<lang::XMultiServiceFactory>();
}

bool OfficeWidget::createFrame(void)
{
  Reference<lang::XMultiServiceFactory> rServiceManager = serviceManager();
  if (!rServiceManager.is())
  {
    kdError() << "Not connected to service manager!" << endl;
    return false;
  }
  
  // Frame (there are two ways to obtain it)
  m_frame = Reference<frame::XFrame>(
      rServiceManager->createInstance(OUSTR("com.sun.star.frame.Task")),
      uno::UNO_QUERY);
  if (!m_frame.is())
  {
    m_frame = Reference<frame::XFrame>(
	rServiceManager->createInstance(OUSTR("com.sun.star.frame.Frame")),
	uno::UNO_QUERY);
  }
  
  // Toolkit
  Reference<uno::XInterface> rToolkitInstance = rServiceManager->createInstance(OUSTR("com.sun.star.awt.Toolkit"));
  Reference<awt::XToolkit> rToolkit(rToolkitInstance, uno::UNO_QUERY);

  // Peer
  Reference<awt::XSystemChildFactory> rChildFactory(rToolkitInstance, uno::UNO_QUERY);
  Any anyParent;
  anyParent <<= winId();
  Sequence<sal_Int8> seqProcessId(1);
  seqProcessId[0] = 0;
  Reference<awt::XWindowPeer> rPeer = rChildFactory->createSystemChild(anyParent,
      seqProcessId, lang::SystemDependent::SYSTEM_XWINDOW);
  
  // Window
  Reference<awt::XWindow> rWindow(rPeer, uno::UNO_QUERY);
  m_frame->initialize(rWindow);

  // Insert the new frame to the desktop hierarchy
  Reference<frame::XFramesSupplier> rTreeRoot(
      rServiceManager->createInstance(OUSTR("com.sun.star.frame.Desktop")),
      uno::UNO_QUERY);
  Reference<frame::XFrames> rChildContainer = rTreeRoot->getFrames();
  rChildContainer->append(m_frame);
  
  // Make it visible
  rWindow->setVisible(true);
  m_frame->setName(getUniqueFrameName());

  kdDebug() << "Frame created." << endl;
  
  return true;
}

bool OfficeWidget::loadDocument(const OUString &name)
{
  if (!m_frame.is() && !createFrame())
  {
    kdError() << "Cannot create frame." << endl;
    return false;
  }

  Reference<frame::XComponentLoader> rComponentLoader(
      serviceManager()->createInstance(OUSTR("com.sun.star.frame.Desktop")),
      uno::UNO_QUERY);
  
  //Sequence<beans::PropertyValue> seqProperties;
  Sequence<beans::PropertyValue> seqProperties(1);
  beans::PropertyValue *arrProps = seqProperties.getArray();
  arrProps[0].Name = OUSTR("ReadOnly");
  arrProps[0].Value <<= sal_True;

  rComponentLoader->loadComponentFromURL(name,
      m_frame->getName(),
      frame::FrameSearchFlag::CHILDREN,
      seqProperties);
  kdDebug() << "Load document: Done." << endl;

  // Do not show menus etc.
  hideOfficeControls();
  
  return true;
}

bool OfficeWidget::hideOfficeControls(void)
{
  bool result = true;
  
  result = hideOfficeControl(SID_TOGGLE_MENUBAR,      "MenuBarVisible")?       result: false;
  result = hideOfficeControl(SID_TOGGLECOMMONTASKBAR, "CommonTaskBarVisible")? result: false;
  result = hideOfficeControl(SID_TOGGLEFUNCTIONBAR,   "FunctionBarVisible")?   result: false;
  result = hideOfficeControl(SID_TOGGLEINETBAR,       "InetBarVisible")?       result: false;
  result = hideOfficeControl(SID_TOGGLEMACROBAR,      "MacroBarVisible")?      result: false;
  result = hideOfficeControl(SID_TOGGLENAVBAR,        "NavigationBarVisible")? result: false;
  result = hideOfficeControl(SID_TOGGLEOBJECTBAR,     "ObjectBarVisible")?     result: false;
  result = hideOfficeControl(SID_TOGGLEOPTIONBAR,     "OptionBarVisible")?     result: false;
  result = hideOfficeControl(SID_TOGGLESTATUSBAR,     "StatusBarVisible")?     result: false;
  result = hideOfficeControl(SID_TOGGLETOOLBAR,       "ToolBarVisible")?       result: false;

  return result;
}

bool OfficeWidget::hideOfficeControl(int slot_no, const char *name)
{
  OUStringBuffer buf(16);
  buf.appendAscii("slot:");
  buf.append((sal_Int32) slot_no);

  Sequence<beans::PropertyValue> seqProps(1);
  beans::PropertyValue *arrProps = seqProps.getArray();
  arrProps[0].Name = OUString::createFromAscii(name);
  arrProps[0].Value <<= sal_False;

  return dispatchUrl(buf.makeStringAndClear(), seqProps);
}

bool OfficeWidget::dispatchUrl(const OUString &url, const Sequence<beans::PropertyValue> &seqProps)
{
  Reference<lang::XMultiServiceFactory> rServiceManager = serviceManager();
  if (!rServiceManager.is())
  {
    kdError() << "Not connected to service manager!" << endl;
    return false;
  }

  if (!m_frame.is())
    return false;

  Reference<util::XURLTransformer> rURLTransformer(
      rServiceManager->createInstance(OUSTR("com.sun.star.util.URLTransformer")),
      uno::UNO_QUERY);
  if (!rURLTransformer.is())
  {
    kdDebug() << "Cannot instantiate XURLTransformer." << endl;
    return false;
  }
  
  util::URL aURL;
  aURL.Complete = url;
  rURLTransformer->parseStrict(aURL);

  Reference<frame::XDispatchProvider> rDispatchProvider(m_frame, uno::UNO_QUERY);
  Reference<frame::XDispatch> rDispatch = rDispatchProvider->queryDispatch(aURL,
      OUString::createFromAscii(""), 0);

  if (!rDispatch.is())
  {
    kdDebug() << "Cannot instantiate XDispatch." << endl;
    return false;
  }
  rDispatch->dispatch(aURL, seqProps);

  return true;
}

OUString OfficeWidget::getUniqueFrameName(void)
{
  OUString  ouBaseName(RTL_CONSTASCII_USTRINGPARAM("kooo-frame-"));
  sal_Int32 lenBaseName = ouBaseName.getLength();
  
  OUStringBuffer ouBuf(25);
  ouBuf.append(ouBaseName, lenBaseName);
  
  Reference<container::XIndexAccess> rContainer;
  try
  {
    Reference<lang::XMultiServiceFactory> rServiceManager = serviceManager();
    if (!rServiceManager.is())
    {
      kdError() << "Not connected to service manager!" << endl;
      ouBuf.appendAscii("12345");
      return ouBuf.makeStringAndClear();
    }
    Reference<frame::XFramesSupplier> rSupplier(
	rServiceManager->createInstance(OUSTR("com.sun.star.frame.Desktop")),
	uno::UNO_QUERY);
    rContainer = Reference<container::XIndexAccess>(
	rSupplier->getFrames(),
	uno::UNO_QUERY);
  }
  catch (Exception &e)
  {
    OString o = OUStringToOString(e.Message, RTL_TEXTENCODING_ASCII_US);
    kdDebug() << "Unable to get a unique name: " << o.pData->buffer << endl;
    ouBuf.appendAscii("12345");
    return ouBuf.makeStringAndClear();
  }
  catch (...)
  {
    kdDebug() << "Unable to get a unique name." << endl;
    ouBuf.appendAscii("12345");
    return ouBuf.makeStringAndClear();
  }

  sal_Int32 no_frames = rContainer->getCount();

  // if we have rContainer->getCount() names, there will surely be at least
  // one unused among (rContainer->getCount() + 1)
  sal_Int32 no_idxs =  no_frames + 1;
  bool *idxs = new bool[no_idxs];
  memset(idxs, false, sizeof(bool)*no_idxs);

  try
  {
    // fill the idxs array
    // (for every 'ouBaseName-index' fill idxs[index] with true)
    for (sal_Int32 i = 0; i < no_frames; ++i)
    {
      uno::Any aItem = rContainer->getByIndex(i);
      Reference<frame::XFrame> rFrame;
      aItem >>= rFrame;
      if (rFrame.is())
      {
	OUString ouFrameName = rFrame->getName();
	//kdDebug() << "Used name: " << OUString2Q(ouFrameName) << endl;

	// we test just frames with the ouBaseName prefx
	if (0 == ouFrameName.compareTo(ouBaseName, lenBaseName))
	{
	  sal_Int32 index = ouFrameName.copy(lenBaseName).toInt32();
	  if (0 <= index && index < no_idxs)
	    idxs[index] = true;
	}
      }
    }
  }
  catch (Exception &e)
  {
    delete[] idxs;
    
    OString o = OUStringToOString(e.Message, RTL_TEXTENCODING_ASCII_US);
    kdDebug() << "Unable to get a unique name: " << o.pData->buffer << endl;
    ouBuf.appendAscii("12345");
    return ouBuf.makeStringAndClear();
  }
  catch (...)
  {
    delete[] idxs;

    kdDebug() << "Unable to get a unique name." << endl;
    ouBuf.appendAscii("12345");
    return ouBuf.makeStringAndClear();
  }

  // and now we can get an unused index
  sal_Int32 index = 0;
  while ((index < no_idxs) && idxs[index])
    ++index;
  delete[] idxs;

  // construct the name (no additinal test needed, see above)
  ouBuf.append(index);

  return ouBuf.makeStringAndClear();
}
